@@ -85,12 +85,11 @@
 //! proof of publication medium on which the seals are defined.
 //!
 //! The module provides two options of implementing sch medium: synchronous
-//! [`SealMedium`] and asynchronous [`AsyncSealMedium`].
+//! [`SealMedium`] and asynchronous `AsyncSealMedium`.
 //!
 //! ## Sample implementation
 //!
-//! Examples of implementations can be found in [`bp::seals`][crate::bp::seals]
-//! module of the crate source code.
+//! Examples of implementations can be found in `bp::seals`.
 //!
 //! ## Further reading
 //!
@@ -103,6 +102,9 @@
 
 #[macro_use]
 extern crate amplify_derive;
+#[cfg(feature = "async")]
+#[macro_use]
+extern crate async_trait;
 
 /// Single-use-seal trait: implement for a data structure that will hold a
 /// single-use-seal definition and will contain a business logic for closing
@@ -111,16 +113,25 @@ extern crate amplify_derive;
 ///
 /// NB: It is recommended that single-use-seal instances to be instantiated
 /// not by a constructor, but by a factory, i.e. "seal medium": data type
-/// implementing either [`SealMedium`] or [`AsyncSealMedium`] traits.
+/// implementing either [`SealMedium`] or `AsyncSealMedium` traits.
 #[cfg_attr(feature = "async", async_trait)]
 pub trait SingleUseSeal {
     /// Associated type for the witness produced by the single-use-seal close
     /// procedure
+    #[cfg(not(feature = "async"))]
     type Witness;
 
+    /// Associated type for the witness produced by the single-use-seal close
+    /// procedure
+    #[cfg(feature = "async")]
+    type Witness: Sync + Send;
+
     /// Type that contains seal definition
     type Definition;
 
+    /// Message type that is supported by the current single-use-seal
+    type Message: AsRef<[u8]>;
+
     /// Closing and verification errors
     type Error: std::error::Error;
 
@@ -129,17 +140,14 @@ pub trait SingleUseSeal {
     /// NB: Closing of the seal MUST not change the internal state of the
     /// seal itself; all the data produced by the process must be placed
     /// into the returned Witness type
-    fn close(
-        &self,
-        over: impl AsRef<[u8]>,
-    ) -> Result<Self::Witness, Self::Error>;
+    fn close(&self, over: &Self::Message)
+        -> Result<Self::Witness, Self::Error>;
 
     /// Verifies that the seal was indeed closed over the message on the
     /// specific seal medium (see [`SealMedium`])
-    #[cfg(not(feature = "async"))]
     fn verify(
         &self,
-        msg: impl AsRef<[u8]>,
+        msg: &Self::Message,
         witness: &Self::Witness,
         medium: &impl SealMedium<Self>,
     ) -> Result<bool, Self::Error>
@@ -149,14 +157,14 @@ pub trait SingleUseSeal {
     /// Verifies that the seal was indeed closed over the message on the
     /// specific seal medium (see [`SealMedium`])
     #[cfg(feature = "async")]
-    async fn verify(
+    async fn verify_async(
         &self,
-        msg: impl AsRef<[u8]>,
+        msg: &Self::Message,
         witness: &Self::Witness,
-        medium: &impl SealMedium<Self>,
+        medium: &impl SealMediumAsync<Self>,
     ) -> Result<bool, Self::Error>
     where
-        Self: Sized;
+        Self: Sized + Sync + Send;
 }
 
 /// Trait for proof-of-publication medium on which the seals are defined and
@@ -168,7 +176,7 @@ pub trait SingleUseSeal {
 ///
 /// Since the medium may require network communications or extensive computing
 /// involved (like in case with blockchain) there is a special asynchronous
-/// version of the SealMedium [`AsyncSealMedium`], which requires use of
+/// version of the seal medium `AsyncSealMedium`, which requires use of
 /// `async` feature of this crate.
 ///
 /// All these operations are medium-specific; for the same single-use-seal type
@@ -176,7 +184,6 @@ pub trait SingleUseSeal {
 ///
 /// To read more on proof-of-publication please check
 /// <https://petertodd.org/2014/setting-the-record-proof-of-publication>
-#[cfg(not(feature = "async"))]
 pub trait SealMedium<Seal>
 where
     Seal: SingleUseSeal,
@@ -209,10 +216,10 @@ where
         Err(SealMediumError::PublicationIdNotSupported)
     }
 
-    /// Returns [Self::PublicationId] for a given witness, if any; the id is
+    /// Returns [`Self::PublicationId`] for a given witness, if any; the id is
     /// returned as an option. Function has default implementation doing
-    /// nothing and just returning [SealMediumError::PublicationIdNotSupported]
-    /// error.
+    /// nothing and just returning
+    /// [`SealMediumError::PublicationIdNotSupported`] error.
     fn get_witness_publication_id(
         &self,
         _witness: &Seal::Witness,
@@ -222,7 +229,7 @@ where
 
     /// Validates whether a given publication id is present in the medium.
     /// Function has default implementation doing nothing and returning
-    /// [SealMediumError::PublicationIdNotSupported] error.
+    /// [`SealMediumError::PublicationIdNotSupported`] error.
     fn validate_publication_id(
         &self,
         _publication_id: &Self::PublicationId,
@@ -234,11 +241,12 @@ where
 /// Asynchronous version of the [`SealMedium`] trait.
 #[cfg(feature = "async")]
 #[async_trait]
-pub trait SealMedium<Seal>
+pub trait SealMediumAsync<Seal>
 where
     Seal: SingleUseSeal + Sync + Send,
     Seal::Witness: Sync + Send,
     Self::PublicationId: Sync,
+    Self: Send + Sync,
 {
     /// Publication id that may be used for referencing publication of
     /// witness data in the medium. By default set `()`, so [`SealMedium`]
@@ -301,7 +309,7 @@ where
 }
 
 /// Single-use-seal status returned by [`SealMedium::get_seal_status`] and
-/// [`AsyncSealMedium::get_seal_status`] functions.
+/// `AsyncSealMedium::get_seal_status` functions.
 ///
 /// NB: It's important to note, that while its possible to deterministically
 ///   define was a given seal closed it yet may be not possible to find out
@@ -322,13 +330,13 @@ pub enum SealStatus {
     Closed = 1,
 }
 
-/// Error returned by [`SealMedium`] and [`AsyncSealMedium`] functions related
+/// Error returned by [`SealMedium`] and `AsyncSealMedium` functions related
 /// to work with publication id ([`SealMedium::PublicationId`]). Required since
-/// not all implementation of [`SealMedia`] may define publication identifier,
+/// not all implementation of [`SealMedium`] may define publication identifier,
 /// and the traits provide default implementation for these functions always
-/// returning [`SealMediumError::OperationNotSupported`]. If the implementation
-/// would like to provide custom implementation, it may embed standard error
-/// related to [`SealMedium`] operations within
+/// returning [`SealMediumError::PublicationIdNotSupported`]. If the
+/// implementation would like to provide custom implementation, it may embed
+/// standard error related to [`SealMedium`] operations within
 /// [`SealMediumError::MediumAccessError`] case; the type of MediumAccessError
 /// is defined through generic argument to [`SealMediumError`].
 #[derive(Clone, Copy, Debug, Display, Error, From)]
